# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121._perf_calc import get_sample_duration
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    PERCEIVED_WAVELENGTH,
    AlgoConfigBase,
    Controller,
    select_prf,
)

from ._processors import Processor, ProcessorConfig


SPARSE_IQ_PPC = 24

NUM_SEGMENTS = 4  # should be a power of 2


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attributes_doc
@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_point: int = attrs.field(default=200)
    """Start point of measurement interval."""

    num_points: int = attrs.field(default=1)
    """Number of points to measure."""

    step_length: Optional[int] = attrs.field(
        default=None,
        validator=attrs.validators.optional(a121.SubsweepConfig.step_length_validator),
    )
    """Step length between points."""

    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    """
    Sets the profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR.
    """
    frame_rate: Optional[float] = attrs.field(default=None)
    """Frame rate in Hz."""

    num_bins: int = attrs.field(default=50)
    """Determines the resolution in m/s by max_speed/num_bins."""

    sweep_rate: Optional[int] = attrs.field(default=None)
    """Sweep rate in Hz."""

    hwaas: Optional[int] = attrs.field(default=None)
    """Number of HWAAS."""

    max_speed: float = attrs.field(default=10.0)
    """Max detectable speed in m/s."""

    threshold: float = attrs.field(default=100.0)
    """
    Peak relative height to median scaled PSD.
    E.g., 10.0 indicates that we need to have 10 times the median value to trigger.
    """

    @classmethod
    def _get_min_sweep_rate(cls, max_speed: float) -> int:
        """
        Calculates the min sweep rate in Hz based on the desired max speed.
        Inverse to _get_max_speed
        """
        # max speed is half half wavelength over sample time (1/sample frequency).
        # Oversample by 10%
        min_sample_freq = 1.1 * 2 * max_speed / PERCEIVED_WAVELENGTH

        sample_freq = int(np.ceil(min_sample_freq))

        return sample_freq

    @classmethod
    def _get_max_speed(cls, sweep_rate: int) -> float:
        """
        Calculates the max speed in m/s based on the desired sweep_rate.
        Inverse to _get_min_sweep_rate
        """
        max_speed = sweep_rate * PERCEIVED_WAVELENGTH / (1.1 * 2)

        return max_speed

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        min_sweep_rate = self._get_min_sweep_rate(self.max_speed)
        frame_rate = self.frame_rate
        sweep_rate = self.sweep_rate
        sweeps_per_frame = self.num_bins * NUM_SEGMENTS

        max_frame_rate = None
        profile_3_direct_leakage_end = int(
            2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_3] / PERCEIVED_WAVELENGTH
        )

        if self.start_point < profile_3_direct_leakage_end:
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "start_point",
                    "Range includes direct leakage, risk for missed detection",
                )
            )

        if self.num_points == 1 and self.step_length is not None:
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "step_length",
                    "Step length is not used when sampling a single point.",
                )
            )

        if self.num_points > 1:
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "num_points",
                    "Several depths measured, " "measures of different objects might be reported.",
                )
            )

        if self.profile == a121.Profile.PROFILE_1 or self.profile == a121.Profile.PROFILE_2:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "profile",
                    "Only Profile 3,4 or 5 is supported.",
                )
            )

        if sweep_rate is not None:
            max_frame_rate = sweep_rate / sweeps_per_frame
            if sweep_rate < min_sweep_rate:
                validation_results.append(
                    a121.ValidationError(
                        self,
                        "sweep_rate",
                        "Max speed measurement too high with set sweep rate."
                        " Increase sweep rate or lower max speed.",
                    )
                )

        if frame_rate is not None and max_frame_rate is not None and frame_rate > max_frame_rate:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "frame_rate",
                    f"Frame rate is too high, max frame rate is approx {max_frame_rate}",
                )
            )

        # Figure out buffer size and warn about that.

        buffer_size = self.num_points * sweeps_per_frame
        if buffer_size > 4095:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "num_bins",
                    f"Buffer size too large ({buffer_size}). Reduce num points or num_bins.",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class DetectorMetadata:
    start_m: float = attrs.field()
    """Actual start point of measurement in meters"""

    step_length_m: float = attrs.field()
    """Actual step length between each data point of the measurement in meters"""

    num_points: int = attrs.field()
    """The number of data points in the measurement"""

    profile: a121.Profile = attrs.field()
    """Profile used for measurement"""


@attrs.frozen(kw_only=True)
class DetectorExtraResult:
    psd: npt.NDArray[np.float64]
    """Full Power Spectral Density from the DFT"""

    velocities: npt.NDArray[np.float64]
    """The frequency bins interpreted as speeds"""

    actual_thresholds: npt.NDArray[np.float64]
    """The thresholds that was used in this frame"""


@attrs.frozen(kw_only=True)
class DetectorResult:
    speed_per_depth: npt.NDArray[np.float64]
    """The measured speed for each depth"""

    extra_result: DetectorExtraResult
    """Detailed results, used for plotting"""

    @property
    def max_speed(self) -> np.float64:
        return max(np.min(self.speed_per_depth), np.max(self.speed_per_depth), key=np.abs)


class Detector(Controller[DetectorConfig, DetectorResult]):
    MIN_DIST_M = {
        a121.Profile.PROFILE_3: None,
        a121.Profile.PROFILE_4: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_4],
        a121.Profile.PROFILE_5: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_5],
    }

    DEFAULT_STEP_LENGTH = {
        a121.Profile.PROFILE_1: 12,
        a121.Profile.PROFILE_2: 24,
        a121.Profile.PROFILE_3: 48,
        a121.Profile.PROFILE_4: 72,
        a121.Profile.PROFILE_5: 120,
    }

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        detector_config: DetectorConfig,
    ) -> None:
        super().__init__(client=client, config=detector_config)
        self.sensor_id = sensor_id
        self.detector_metadata: Optional[DetectorMetadata] = None

        self.started = False
        self.config = detector_config
        self.timing = 0.0

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        self.processor_config = self._get_processor_config(self.config)
        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=self.processor_config,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("speed_detector")
                _record_algo_data(
                    _algo_group,
                    self.sensor_id,
                    self.config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")
            self.client.attach_recorder(recorder)

        self.detector_metadata = DetectorMetadata(
            start_m=sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M,
            step_length_m=metadata.base_step_length_m * sensor_config.step_length,
            num_points=sensor_config.num_points,
            profile=sensor_config.profile,
        )
        self.client.start_session()

        self.started = True

    @classmethod
    def _get_max_hwaas(
        cls, sweep_rate: int, num_points: int, prf: a121.PRF, profile: a121.Profile
    ) -> int:
        sample_duration = get_sample_duration(prf=prf, profile=profile)

        MAGIC_OVERHEAD_S = 2 * 1e-6  # Constant added in rss "for good measure"

        # 0.98 is also an added percentage from rss that is accounted for here
        max_sweep_duration = (
            ((1.0 / sweep_rate) - MAGIC_OVERHEAD_S) * 0.98 / (num_points * sample_duration)
        )
        max_hwaas = int(np.floor(max_sweep_duration - 4))  # all overhead duration is at most 4

        max_hwaas = min(max_hwaas, a121.SubsweepConfig.MAX_HWAAS)

        return max_hwaas

    @classmethod
    def _get_step_length(cls, profile: a121.Profile) -> int:
        """
        Returns a minimal step length in order to achieve independent points,
        based on profile.
        """
        return cls.DEFAULT_STEP_LENGTH[profile]

    @classmethod
    def _get_sensor_config(cls, config: DetectorConfig) -> a121.SensorConfig:
        start_m = config.start_point * APPROX_BASE_STEP_LENGTH_M

        if config.profile is not None:
            profile = config.profile
        else:
            # Take the highest possible profile
            viable_profiles = [k for k, v in cls.MIN_DIST_M.items() if v is None or v <= start_m]
            profile = viable_profiles[-1]

        if config.sweep_rate is not None:
            sweep_rate = config.sweep_rate
        else:
            sweep_rate = config._get_min_sweep_rate(config.max_speed)

        frame_rate = config.frame_rate  # This can be none

        if config.step_length is not None:
            step_length = config.step_length
        else:
            step_length = cls._get_step_length(profile)

        end_point = config.start_point + (config.num_points - 1) * step_length
        prf = select_prf(end_point, profile)

        if config.hwaas is not None:
            hwaas = config.hwaas
        else:
            max_hwaas = cls._get_max_hwaas(sweep_rate, config.num_points, prf, profile)
            hwaas = max(int(max_hwaas * 0.9), 1)

        sensor_config = a121.SensorConfig(
            profile=profile,
            start_point=config.start_point,
            num_points=config.num_points,
            step_length=step_length,
            prf=prf,
            hwaas=hwaas,
            sweeps_per_frame=NUM_SEGMENTS * config.num_bins,
            sweep_rate=sweep_rate,
            frame_rate=frame_rate,
            inter_frame_idle_state=a121.IdleState.READY,
            double_buffering=False,
            continuous_sweep_mode=False,
        )
        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: DetectorConfig) -> ProcessorConfig:
        ret_config = ProcessorConfig(threshold=config.threshold, num_segments=NUM_SEGMENTS)
        return ret_config

    def get_next(self) -> DetectorResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        extra_result = DetectorExtraResult(
            psd=processor_result.extra_result.psd,
            velocities=processor_result.extra_result.velocities,
            actual_thresholds=processor_result.extra_result.actual_thresholds,
        )

        detector_result = DetectorResult(
            extra_result=extra_result,
            speed_per_depth=processor_result.speed_per_depth,
        )
        return detector_result

    def update_config(self, config: DetectorConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        recorder_result = self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()
        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: DetectorConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, DetectorConfig]:
    sensor_id = algo_group["sensor_id"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])
    return sensor_id, config



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from acconeer.exptool._core.communication.client import ClientError, ServerError

from .client import Client
from .exploration_client import ExplorationClient
from .exploration_protocol import (
    ExplorationProtocol,
    ExplorationProtocolError,
    get_exploration_protocol,
)
from .mock_client import MockClient



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from ._aggregator import (
    Aggregator,
    AggregatorConfig,
    AggregatorResult,
    PeakSortingMethod,
    ProcessorSpec,
)
from ._context import DetectorContext
from ._detector import (
    PRF_REMOVED_ET_VERSION,
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorResult,
    _DetectorConfig_v0,
)
from ._processors import (
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ProcessorResult,
    ReflectorShape,
    ThresholdMethod,
    calculate_bg_noise_std,
)
from ._translation import detector_config_to_processor_specs, detector_config_to_session_config



# Copyright (c) Acconeer AB, 2022-2026
# All rights reserved
from __future__ import annotations

import copy
import typing as t
from typing import Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase
from acconeer.exptool.a121.algo.distance import DetectorConfig, DetectorContext, DetectorResult


@attrs.frozen(kw_only=True)
class Point:
    """A point located in the 2D plane spanned by the two sensors."""

    angle: float
    distance: float
    x_coord: float
    y_coord: float


@attrs.frozen(kw_only=True)
class ObjectWithoutCounterpart:
    """Object that was not matched with distance from other sensor."""

    distance: float
    sensor_position: str


@attrs.frozen(kw_only=True)
class _KalmanFilterResult:
    """Result from a Kalman filter."""

    distance: float
    sensor_position: str


@attrs.frozen(kw_only=True)
class ProcessorResult:
    """Processor result"""

    points: t.List[Point]
    objects_without_counterpart: t.List[ObjectWithoutCounterpart]


@attributes_doc
@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    sensor_spacing_m: float = attrs.field(default=0.1)
    """Distance between the two sensors (m) in the intended installation."""
    max_anticipated_velocity_mps: float = attrs.field(default=2.0)
    """Specify the highest anticipated radial velocity between sensors/device and
    objects (m/s)."""
    dead_reckoning_duration_s: float = attrs.field(default=0.5)
    """Specify the duration (s) of the Kalman filter to perform dead reckoning before stop
    tracking."""
    sensitivity: float = attrs.field(default=0.5)
    """Specify the sensitivity of the Kalman filter. A higher value yields a more responsive
    filter. A lower value yields a more robust filter."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.update_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config,
                    "update_rate",
                    "Must be set",
                )
            )

        return validation_results


class Processor:
    """Bilateration processor
    :param sensor_config: Sensor configuration.
    :param extended_metadata: Metadata yielded by the sensor config.
    :param processor_config: Processor configuration.
    :param sensor_ids: Sensor IDs used.
    """

    _SENSOR_POSITION_LEFT = "left"
    _SENSOR_POSITION_RIGHT = "right"
    _MAX_NUM_OBJECTS = 10

    def __init__(
        self,
        session_config: a121.SessionConfig,
        processor_config: ProcessorConfig,
        sensor_ids: t.List[int],
    ):
        if len(sensor_ids) != 2:
            msg = "Number of sensor ids must equal two."
            raise ValueError(msg)
        processor_config.validate(session_config)
        assert session_config.update_rate is not None  # Should never happen, checked in validate
        self.update_rate = session_config.update_rate
        self.sensor_spacing_m = processor_config.sensor_spacing_m
        self.process_noise_gain_sensitivity = processor_config.sensitivity
        self.num_dead_reckoning_frames = int(
            processor_config.dead_reckoning_duration_s * self.update_rate
        )
        # Largest anticipated distance change of target between frames. 4 to add margin.
        self.max_meas_state_diff_m = (
            processor_config.max_anticipated_velocity_mps / self.update_rate * 4
        )
        self.min_num_updates_valid_estimate = self._sensitivity_to_min_num_updates_for_tracking(
            processor_config.sensitivity
        )
        self.sensor_position_to_ids = {
            self._SENSOR_POSITION_LEFT: sensor_ids[0],
            self._SENSOR_POSITION_RIGHT: sensor_ids[1],
        }
        self.left_sensor_kfs: list[_KalmanFilter] = []
        self.right_sensor_kfs: list[_KalmanFilter] = []

    def process(self, result: t.Dict[int, DetectorResult]) -> ProcessorResult:
        distances_left = result[self.sensor_position_to_ids[self._SENSOR_POSITION_LEFT]].distances
        rcs_left = result[self.sensor_position_to_ids[self._SENSOR_POSITION_LEFT]].strengths
        distances_right = result[
            self.sensor_position_to_ids[self._SENSOR_POSITION_RIGHT]
        ].distances
        rcs_right = result[self.sensor_position_to_ids[self._SENSOR_POSITION_RIGHT]].strengths
        assert distances_left is not None
        assert rcs_left is not None
        assert distances_right is not None
        assert rcs_right is not None
        # Get distances from detector result and remove closely spaced distances.
        distances_left_cleaned = self._remove_closely_spaced_distances(
            distances_left,
            rcs_left,
            self.sensor_spacing_m,
        )
        distances_right_cleaned = self._remove_closely_spaced_distances(
            distances_right,
            rcs_right,
            self.sensor_spacing_m,
        )
        # Truncate list to a known max length.
        if len(distances_left_cleaned) > self._MAX_NUM_OBJECTS:
            distances_left_cleaned = distances_left_cleaned[self._MAX_NUM_OBJECTS :]
        if len(distances_right_cleaned) > self._MAX_NUM_OBJECTS:
            distances_right_cleaned = distances_right_cleaned[self._MAX_NUM_OBJECTS :]
        # Update kalman filters.
        self.left_sensor_kfs = self._update_kalman_filters(
            self.left_sensor_kfs, distances_left_cleaned, self._SENSOR_POSITION_LEFT
        )
        self.right_sensor_kfs = self._update_kalman_filters(
            self.right_sensor_kfs,
            distances_right_cleaned,
            self._SENSOR_POSITION_RIGHT,
        )
        # Extract result after filtering.
        kf_result_left_sensor = [
            _KalmanFilterResult(distance=kf.get_distance(), sensor_position=kf.sensor_position)
            for kf in self.left_sensor_kfs
            if kf.has_init
        ]
        kf_result_right_sensor = [
            _KalmanFilterResult(distance=kf.get_distance(), sensor_position=kf.sensor_position)
            for kf in self.right_sensor_kfs
            if kf.has_init
        ]
        # Match result from both sensors to create pairs and objects without counterpart.
        (points, objects_without_counterpart) = self._pair_distances(
            kf_result_left_sensor, kf_result_right_sensor, self.sensor_spacing_m
        )
        return ProcessorResult(
            points=points, objects_without_counterpart=objects_without_counterpart
        )

    def _pair_distances(
        self,
        kf_result_left_sensor: t.List[_KalmanFilterResult],
        kf_result_right_sensor: t.List[_KalmanFilterResult],
        sensor_spacing: float,
    ) -> t.Tuple[t.List[Point], t.List[ObjectWithoutCounterpart]]:
        """Pair distance from each sensor to form points.
        The sensor with the least number of results is identified and looped over. In each loop,
        the distance is matched to a distance from the other sensor. The condition for a match is
        the absolute distance difference being lower than the sensor spacing.
        A distance can only be matched once. When matched, it is marked and can not be
        matched other distances.
        Each pair is used to form a point, for which the distance and angle is calculated, along
        with its cartesian coordinates.
        Distances without a pair is regarded as an object without a counterpart.

        The bilateration function expects the distance estimate from the left sensor as the first
        argmument. If the value from the right sensor is fed as the first element, the sign of the
        angle needs to be flipped.
        """
        if len(kf_result_left_sensor) <= len(kf_result_right_sensor):
            shorter_result = kf_result_left_sensor
            longer_result = kf_result_right_sensor
            flip_angle = False
        else:
            shorter_result = kf_result_right_sensor
            longer_result = kf_result_left_sensor
            flip_angle = True

        shorter_result_item_has_pair = [False] * len(shorter_result)
        longer_result_item_has_pair = [False] * len(longer_result)
        distances_longer_result = np.array([kf_result.distance for kf_result in longer_result])
        points = []
        for kf_idx, kf_shorter in enumerate(shorter_result):
            # Find the closest distance in the other array
            idx_closest = np.argmin(np.abs(kf_shorter.distance - distances_longer_result))
            # Add as a pair, if the distance is within the expected range(plus a small margin).
            if np.abs(kf_shorter.distance - distances_longer_result[idx_closest]) < sensor_spacing:
                distance = (kf_shorter.distance + distances_longer_result[idx_closest]) / 2

                angle = self._estimate_angle(
                    kf_shorter.distance, distances_longer_result[idx_closest], sensor_spacing
                )

                if flip_angle:
                    angle = -angle

                points.append(
                    Point(
                        angle=angle,
                        distance=distance,
                        x_coord=np.sin(angle) * distance,
                        y_coord=np.cos(angle) * distance,
                    )
                )
                shorter_result_item_has_pair[kf_idx] = True
                longer_result_item_has_pair[idx_closest] = True
        objects_without_counterpart = []
        for i in [i for i, has_init in enumerate(shorter_result_item_has_pair) if not has_init]:
            objects_without_counterpart.append(
                ObjectWithoutCounterpart(
                    distance=shorter_result[i].distance,
                    sensor_position=shorter_result[i].sensor_position,
                )
            )
        for i in [i for i, has_init in enumerate(longer_result_item_has_pair) if not has_init]:
            objects_without_counterpart.append(
                ObjectWithoutCounterpart(
                    distance=longer_result[i].distance,
                    sensor_position=longer_result[i].sensor_position,
                )
            )
        return (points, objects_without_counterpart)

    @staticmethod
    def _remove_closely_spaced_distances(
        distances: npt.NDArray[np.float64], rcs: npt.NDArray[np.float64], min_dist: float
    ) -> t.List[float]:
        """Remove closely spaced distances to avoid ambiguity in later filtering and distance
        pairing stages.
        If two distance are closely spaced, keep the one with highest RCS.
        """
        if len(distances) == 0:
            return []
        # Sort according to distance.
        distances, rcs = zip(*sorted(zip(distances, rcs)))  # type: ignore[assignment]
        # If two distance are closer than the senor spacing, remove the one with smaller amplitude.
        index_closely_spaced_distances = np.where(np.abs(np.diff(distances)) < min_dist)[0]
        # Flip the array of indexes to avoid altering the order of the elements to the left in
        # the array.
        for index in np.flip(index_closely_spaced_distances):
            # Remove the element with the lower amplitude.
            if rcs[index] < rcs[index + 1]:
                rcs = np.delete(rcs, index)
                distances = np.delete(distances, index)
            else:
                rcs = np.delete(rcs, index + 1)
                distances = np.delete(distances, index + 1)
        return list(distances)

    def _update_kalman_filters(
        self,
        kfs: t.List[_KalmanFilter],
        distances: t.List[float],
        sensor_position: str,
    ) -> t.List[_KalmanFilter]:
        """Update Kalman filters for a sensor, using new distance estimates.
        Identifying the distance closest to the current state of the filter. If the distance is
        sufficiently close to the current state, it is used to update the filter. Once a distance
        has been used to update a filter, it can't be used by other filters.
        If no estimated distance matches the filter, dead reckoning is used. Each time dead
        reckoning is performed, a counter is incremented. If the counter exceeds a certain value,
        the filter is deleted.
        A filter must have a minimum number of updates before it is regarded as initiated and used
        for bilateration in a subsequent steps.
        """
        distances = copy.deepcopy(list(distances))
        kfs = copy.deepcopy(kfs)
        for kf in kfs:
            # Find the index of the closest distance.
            state_vs_distance_diff = np.abs(np.array(kf.get_distance()) - np.array(distances))
            idxs_close_to_estimate = np.where(state_vs_distance_diff < self.max_meas_state_diff_m)[
                0
            ]
            kf.predict()
            if len(distances) == 0 or len(idxs_close_to_estimate) == 0:
                # No estimates found.
                kf.dead_reckoning_count += 1
                # Remove the filter if not initialized or number of dead reckoning steps is to
                # high.
                if (
                    self.num_dead_reckoning_frames < kf.dead_reckoning_count
                    or kf.num_updates < self.num_dead_reckoning_frames - 1
                ):
                    kfs.remove(kf)
            else:
                # Estimates found.
                kf.dead_reckoning_count = 0
                # Find the point in the data closest to the current estimated distance.
                idx_closest = np.argmin(state_vs_distance_diff[idxs_close_to_estimate])
                idx_in_r_array = idxs_close_to_estimate[idx_closest]
                # Update KF
                kf.update(distances[idx_in_r_array])
                # Remove distance as it has now been used to update a filter.
                distances.pop(idx_in_r_array)
                # Check if the minimum number of updates have been reached. If so, the filter has
                # been initialized.
                if self.min_num_updates_valid_estimate <= kf.num_updates:
                    kf.has_init = True
        for distance in distances:
            kfs.append(
                _KalmanFilter(
                    1 / self.update_rate,
                    self.process_noise_gain_sensitivity,
                    distance,
                    sensor_position,
                    self.min_num_updates_valid_estimate,
                )
            )
        return kfs

    @staticmethod
    def _estimate_angle(
        left_sensor: float, right_sensor: float, sensor_spacing: float
    ) -> float | t.Any:
        """Calculates the angle to an object given two distance values. The first argument should
        reflect the value at the left sensor(left from the perspective of the sensor, facing
        forward). The second argument should reflect the value of the right sensor."""
        if sensor_spacing < np.abs(left_sensor - right_sensor):
            return np.nan
        x0 = left_sensor**2 - right_sensor**2
        x1 = np.sqrt(
            2 * sensor_spacing**2 * (left_sensor**2 + right_sensor**2)
            - (left_sensor**2 - right_sensor**2) ** 2
            - sensor_spacing**4 / 2
        )
        return np.arctan(x0 / x1)

    @staticmethod
    def _sensitivity_to_min_num_updates_for_tracking(sensitivity: float) -> int:
        return int(2 + (1 - sensitivity) * 20)


class _KalmanFilter:
    # Acceleration noise std (m/s^2).
    _PROCESS_NOISE_STD = 0.01
    # Distance estimated noise std (m).
    _MEASUREMENT_NOISE_STD = 0.005

    def __init__(
        self,
        dt: float,
        process_noise_gain_sensitivity: float,
        init_state: float,
        sensor_position: str,
        min_num_updates_valid_estimate: int,
    ) -> None:
        self.A: npt.NDArray[np.float64] = np.matrix([[1.0, dt], [0.0, 1.0]])
        self.H: npt.NDArray[np.float64] = np.matrix([[1, 0]])
        process_noise_gain = self._sensitivity_to_gain(process_noise_gain_sensitivity)
        # Random acceleration process noise.
        self.Q = (
            np.matrix([[(dt**4) / 4, (dt**3) / 2], [(dt**3) / 2, dt**2]])
            * (self._PROCESS_NOISE_STD) ** 2
            * process_noise_gain
        )
        self.R = self._MEASUREMENT_NOISE_STD**2
        self.P = np.eye(self.A.shape[1])
        self.x: npt.NDArray[np.float64] = np.matrix([[init_state], [0.0]])
        self.min_num_updates_valid_estimate = min_num_updates_valid_estimate
        self.dead_reckoning_count = 0
        self.num_updates = 0
        self.sensor_position = sensor_position
        self.has_init = False

    def predict(self) -> None:
        self.x = np.dot(self.A, self.x)
        self.P = np.dot(np.dot(self.A, self.P), self.A.T) + self.Q

    def update(self, z: float) -> None:
        S = np.dot(self.H, np.dot(self.P, self.H.T)) + self.R
        K = np.dot(np.dot(self.P, self.H.T), np.linalg.inv(S))
        self.x = self.x + np.dot(K, (z - np.dot(self.H, self.x)))
        I = np.eye(self.H.shape[1])
        self.P = (I - (K * self.H)) * self.P
        self.num_updates += 1
        if self.min_num_updates_valid_estimate < self.num_updates:
            self.has_init = True

    def get_distance(self) -> float:
        return float(self.x[0, 0])

    @staticmethod
    def _sensitivity_to_gain(sensitivity: float) -> float:
        return 0.01 + sensitivity * 20.0


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[list[int], DetectorConfig, ProcessorConfig, DetectorContext]:
    sensor_ids = algo_group["sensor_ids"][()].tolist()
    detector_config = DetectorConfig.from_json(algo_group["detector_config"][()])
    bilateration_config = ProcessorConfig()

    context_group = algo_group["context"]
    context = opser.deserialize(context_group, DetectorContext)
    return sensor_ids, detector_config, bilateration_config, context



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import logging
import time
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py

from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.presence import DetectorConfig
from acconeer.exptool.a121.algo.presence._pidget_mapping import (
    get_pidget_mapping as _get_presence_pidget_mapping,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget

from ._example_app import ExampleAppConfig
from ._mode_handler import (
    AppMode,
    DetectionState,
    ModeHandler,
    ModeHandlerConfig,
    ModeHandlerResult,
    _load_algo_data,
    get_default_config,
)


opser.register_json_presentable(ModeHandlerConfig)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: ModeHandlerConfig = attrs.field(factory=ModeHandlerConfig)


class PluginPresetId(Enum):
    DEFAULT = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    mode_handler_config: ModeHandlerConfig
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], ModeHandlerConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: get_default_config()
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._exempel_app_instance: Optional[ModeHandler] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = opser.deserialize(file["config"], ModeHandlerConfig)

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: ModeHandlerConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        cfg_group = file.create_group("config")
        opser.serialize(self.shared_state.config, cfg_group)

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.session(0).sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._example_app_instance = ModeHandler(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            mode_handler_config=self.shared_state.config,
        )
        self._example_app_instance.start(recorder)
        self.callback(SetupMessage(mode_handler_config=self.shared_state.config))

    def end_session(self) -> None:
        if self._example_app_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._example_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._example_app_instance is None:
            raise RuntimeError
        result = self._example_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ModeHandlerResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            if not isinstance(message.result, ModeHandlerResult):
                msg = f"Unexpected result type: {type(message.result)}"
                raise RuntimeError(msg)

            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(mode_handler_config=message.mode_handler_config)
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        mode_handler_config: ModeHandlerConfig,
    ) -> None:
        self.plot_layout.clear()

        brush = et.utils.pg_brush_cycler(0)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush, symbolPen="k")

        self.history_plot = self.plot_layout.addPlot()
        self.history_plot.setMenuEnabled(False)
        self.history_plot.showGrid(x=False, y=True)
        self.history_plot.setLabel("left", "Score")
        self.history_plot.setLabel("bottom", "Time (s)")

        self.history_curve = self.history_plot.plot(
            **dict(pen=et.utils.pg_pen_cycler(0), **symbol_dot_kw)
        )

        self.threshold_line = pg.InfiniteLine(pen=et.utils.pg_pen_cycler(1), angle=0)
        self.history_plot.addItem(self.threshold_line)
        self.threshold_line.show()

        self.smooth_max_history = et.utils.SmoothMax(tau_decay=10.0)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:16pt;">'
            "{}</span></div>"
        )
        self.not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:16pt;">'
            "{}</span></div>"
        )
        self.not_present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.present_text_item = pg.TextItem(
            html=self.not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )
        self.present_text_item.hide()
        self.not_present_text_item.hide()
        self.history_plot.addItem(self.present_text_item)
        self.history_plot.addItem(self.not_present_text_item)

        self.history_plot.setXRange(-10, 0)
        self.history_plot.setYRange(0, 5)

    def draw_plot_job(self, example_app_result: ModeHandlerResult) -> None:
        if example_app_result.app_mode == AppMode.PRESENCE:
            self.history_plot.setXRange(-10, 0)
            self.history_plot.setYRange(0, 5)
            self.history_curve.setData([], [])
            num_dots = int(time.time() * 2 % 3) + 1
            present_html = self.not_present_html.format("Monitoring presence" + "." * num_dots)
            self.not_present_text_item.setHtml(present_html)
            self.not_present_text_item.show()
        elif (
            example_app_result.app_mode == AppMode.HANDMOTION
            and example_app_result.example_app_result is not None
        ):
            history = example_app_result.example_app_result.extra_result.history
            history_time = example_app_result.example_app_result.extra_result.history_time
            threshold = example_app_result.example_app_result.extra_result.threshold
            detection_state = example_app_result.detection_state

            lim = max(self.smooth_max_history.update(history), threshold + 1)
            self.history_curve.setData(history_time, history)
            self.history_plot.setYRange(0, lim)
            self.threshold_line.setValue(threshold)

            if detection_state is DetectionState.NO_DETECTION:
                present_html = self.not_present_html.format("No hand detected")
                self.not_present_text_item.setHtml(present_html)
                self.not_present_text_item.show()
                self.present_text_item.hide()
            else:
                if detection_state is DetectionState.DETECTION:
                    present_html = self.present_html_format.format("Hand detected")
                elif detection_state is DetectionState.RETENTION:
                    present_html = self.present_html_format.format("Retaining detection")
                self.present_text_item.setHtml(present_html)
                self.present_text_item.show()
                self.not_present_text_item.hide()

        text_y_pos = self.history_plot.getAxis("left").range[1] * 0.95
        text_x_pos = (
            self.history_plot.getAxis("bottom").range[1]
            + self.history_plot.getAxis("bottom").range[0]
        ) / 2.0

        self.not_present_text_item.setPos(text_x_pos, text_y_pos)
        self.present_text_item.setPos(text_x_pos, text_y_pos)


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor(
            title="Mode handler parameters",
            factory_mapping=self._get_example_app_pidget_mapping(),
            config_type=ModeHandlerConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.setup_editor = AttrsConfigEditor(
            title="Hand motion - Setup parameters",
            factory_mapping=self._get_setup_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.setup_editor.sig_update.connect(self._on_example_app_config_update)
        scrolly_layout.addWidget(self.setup_editor)

        self.filtering_editor = AttrsConfigEditor(
            title="Hand motion - Filtering parameters",
            factory_mapping=self._get_filtering_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.filtering_editor.sig_update.connect(self._on_example_app_config_update)
        scrolly_layout.addWidget(self.filtering_editor)

        self.sensor_config_editor = AttrsConfigEditor(
            title="Hand motion - Sensor configuration parameters",
            factory_mapping=self._get_sensor_config_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.sensor_config_editor.sig_update.connect(self._on_example_app_config_update)
        scrolly_layout.addWidget(self.sensor_config_editor)

        self.presence_config_editor = AttrsConfigEditor(
            title="Presence - Detector parameters",
            factory_mapping=_get_presence_pidget_mapping(),
            config_type=DetectorConfig,
            parent=self.scrolly_widget,
        )
        self.presence_config_editor.sig_update.connect(self._on_presence_config_update)
        self.collapsible_widget = CollapsibleWidget(
            "Presence configuration parameters", self.presence_config_editor, self.scrolly_widget
        )
        scrolly_layout.addWidget(self.collapsible_widget)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_example_app_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "hand_detection_timeout": pidgets.FloatPidgetFactory(
                name_label_text="Hand motion timeout limit:",
                name_label_tooltip=get_attribute_docstring(
                    ModeHandlerConfig, "hand_detection_timeout"
                ),
                limits=(0, None),
                decimals=1,
                suffix="s",
            ),
            "use_presence_detection": pidgets.CheckboxPidgetFactory(
                name_label_text="Enable low-power mode",
                name_label_tooltip=get_attribute_docstring(
                    ModeHandlerConfig, "use_presence_detection"
                ),
            ),
        }

    @classmethod
    def _get_setup_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "sensor_to_water_distance": pidgets.FloatPidgetFactory(
                name_label_text="Distance between sensor and water jet:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "sensor_to_water_distance"
                ),
                limits=(0, 0.3),
                decimals=3,
                suffix="m",
            ),
            "water_jet_width": pidgets.FloatPidgetFactory(
                name_label_text="Water jet width:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "water_jet_width"),
                limits=(0, 0.1),
                decimals=3,
                suffix="m",
            ),
            "measurement_range_end": pidgets.FloatPidgetFactory(
                name_label_text="Measurement range end:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "measurement_range_end"
                ),
                limits=(0, 0.5),
                decimals=3,
                suffix="m",
            ),
        }

    @classmethod
    def _get_filtering_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "threshold": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "threshold"),
                limits=(0, 5),
                decimals=2,
            ),
            "filter_time_const": pidgets.FloatSliderPidgetFactory(
                name_label_text="Filter constant:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "filter_time_const"),
                limits=(0, 2.0),
                decimals=2,
                suffix="s",
            ),
            "detection_retention_duration": pidgets.FloatSliderPidgetFactory(
                name_label_text="Detection retention duration:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "detection_retention_duration"
                ),
                limits=(0, 5),
                decimals=1,
                suffix="s",
            ),
        }

    @classmethod
    def _get_sensor_config_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "sweeps_per_frame": pidgets.IntPidgetFactory(
                name_label_text="Sweeps per frame:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweeps_per_frame"),
                limits=(1, 1000),
            ),
            "sweeps_rate": pidgets.IntPidgetFactory(
                name_label_text="Sweep rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweeps_rate"),
                limits=(1, 5000),
                suffix="Hz",
            ),
            "frame_rate": pidgets.IntPidgetFactory(
                name_label_text="Frame rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "frame_rate"),
                limits=(1, 100),
                suffix="Hz",
            ),
            "hwaas": pidgets.IntPidgetFactory(
                name_label_text="HWAAS:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "hwaas"),
                limits=(1, 511),
            ),
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.sensor_config_editor.set_data(None)
            self.config_editor.set_data(None)
            self.filtering_editor.set_data(None)
            self.setup_editor.set_data(None)
            self.presence_config_editor.set_data(None)
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.config_editor.set_data(state.config)
            self.filtering_editor.set_data(state.config.example_app_config)
            self.setup_editor.set_data(state.config.example_app_config)
            self.sensor_config_editor.set_data(state.config.example_app_config)
            self.presence_config_editor.set_data(state.config.presence_config)

            results = state.config._collect_validation_results()

            not_handled = self.filtering_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.sensor_id_pidget,
                self.config_editor,
                self.filtering_editor,
                self.setup_editor,
                self.sensor_config_editor,
                self.presence_config_editor,
            ],
        )

        configs_valid = (
            self.filtering_editor.is_ready
            and self.sensor_config_editor.is_ready
            and self.setup_editor.is_ready
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=configs_valid)
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: ModeHandlerConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_example_app_config_update(self, example_app_config: ExampleAppConfig) -> None:
        config = self.config_editor.get_data()
        assert config is not None
        config.example_app_config = example_app_config
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_presence_config_update(self, detector_config: DetectorConfig) -> None:
        config = self.config_editor.get_data()
        assert config is not None
        config.presence_config = detector_config
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


HAND_MOTION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="hand_motion",
    title="Hand motion detection",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/hand_motion_detection.html",
    description="Wake-up water faucet application.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import re
import time
from typing import Any, Optional

from PySide6.QtCharts import QChartView, QLineSeries
from PySide6.QtCore import Qt, SignalInstance
from PySide6.QtGui import QPen, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import (
    QAbstractItemView,
    QPushButton,
    QTabWidget,
    QTextEdit,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app.new.app_model import AppModel


class _CurrentStateView(QTextEdit):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()
        app_model.sig_notify.connect(self._update_text)
        self.setFontFamily("monospace")
        self.setReadOnly(True)
        self.update_count = 0

    @staticmethod
    def _stringify_backend_state(backend_state: Optional[Any]) -> list[str]:
        if backend_state is None:
            return ["None"]

        try:
            members = backend_state.__slots__
        except AttributeError:
            members = backend_state.__dict__

        public_members = [member for member in members if not member.startswith("_")]

        return [
            f">>> {attribute}\n" + f"{getattr(backend_state, attribute)}\n"
            for attribute in public_members
        ]

    def _update_text(self, app_model: AppModel) -> None:
        self.update_count += 1
        self.setText(
            "\n".join(
                [
                    "=== AppModel ===",
                    "",
                    f"ConnectionState:          {app_model.connection_state}",
                    f"ConnectionInterface:      {app_model.connection_interface}",
                    f"PluginState:              {app_model.plugin_state}",
                    f"PluginGeneration:         {app_model.plugin_generation}",
                    f"socket_connection_ip:     {app_model.socket_connection_ip}",
                    f"serial_connection_device: {app_model.serial_connection_device}",
                    f"overridden_baudrate:      {app_model.overridden_baudrate}",
                    f"usb_connection_device:    {app_model.usb_connection_device}",
                    f"autoconnect_enabled:      {app_model.autoconnect_enabled}",
                    f"flow_control_enabled:     {app_model.flow_control_enabled}",
                    "",
                    "backend_plugin_state:",
                    *self._stringify_backend_state(app_model.backend_plugin_state),
                    "",
                    f"Update count:             {self.update_count}",
                ]
            )
        )


class _SignalLog(QWidget):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self._model = QStandardItemModel(self)
        self._root = self._model.invisibleRootItem()

        signal_items = (
            self._connect_signal_to_log(app_model.sig_error),
            self._connect_signal_to_log(app_model.sig_load_plugin),
            self._connect_signal_to_log(app_model.sig_message_plot_plugin),
            self._connect_signal_to_log(app_model.sig_message_view_plugin),
            self._connect_signal_to_log(app_model.sig_status_file_path),
            self._connect_signal_to_log(app_model.sig_status_message),
            self._connect_signal_to_log(app_model.sig_rate_stats),
            self._connect_signal_to_log(app_model.sig_frame_count),
            self._connect_signal_to_log(app_model.sig_resource_tab_input_block_requested),
            self._connect_signal_to_log(app_model.sig_backend_state_changed),
        )

        log_view = QTreeView()
        log_view.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        log_view.setHeaderHidden(True)
        log_view.setModel(self._model)

        clear_button = QPushButton("Clear log")

        def remove_logs() -> None:
            for i in signal_items:
                i.removeRows(0, i.rowCount())

        clear_button.pressed.connect(remove_logs)

        layout = QVBoxLayout()
        layout.addWidget(clear_button)
        layout.addWidget(log_view)
        self.setLayout(layout)

    def _connect_signal_to_log(self, signal: SignalInstance) -> QStandardItem:
        pattern = r"<PySide6.QtCore.SignalInstance ([a-zA-Z_]+).*"
        if (match := re.match(pattern, str(signal))) is None:
            signal_name = "unknown signal"
        else:
            (signal_name,) = match.groups("unknown signal")

        item = QStandardItem(signal_name)
        signal.connect(lambda x: item.appendRow(QStandardItem(str(x))))
        self._root.appendRow(item)
        return item


class _TimingPlot(QChartView):
    def __init__(
        self,
        app_model: AppModel,
        capture_len_s: int = 60,
    ) -> None:
        super().__init__()
        self._capture_len_s = capture_len_s
        self._timing_start = time.perf_counter()

        self._app_model = app_model
        self._app_model.sig_timing.connect(self.plot)
        self.reset()

    def reset(self) -> None:
        self._timing_start = time.perf_counter()
        self.chart().removeAllSeries()

        reference_60_hz = QLineSeries()
        reference_60_hz.setName("Reference: 60Hz")
        pen = QPen(Qt.PenStyle.DashLine)
        pen.setColor("red")
        reference_60_hz.setPen(pen)
        reference_60_hz.append(0, 1000 / 60)
        reference_60_hz.append(self._capture_len_s, 1000 / 60)
        self.chart().addSeries(reference_60_hz)

    @property
    def _timing_series(self) -> list[QLineSeries]:
        return [
            s
            for s in self.chart().series()
            if isinstance(s, QLineSeries)
            # if "reference" not in s.name().lower()
        ]

    def plot(self, name_and_times: tuple[str, float, float]) -> None:
        name, start, end = name_and_times
        timestamp = start - self._timing_start

        if timestamp > self._capture_len_s:
            self.reset()
            self.plot(name_and_times)
            return

        duration_ms = 1000 * (end - start)

        timing_series = self._timing_series
        queried_series = next((s for s in timing_series if s.name() == name), None)

        if queried_series is None:
            new_series = QLineSeries()
            new_series.setName(name)
            new_series.append(timestamp, duration_ms)
            self.chart().addSeries(new_series)
        else:
            queried_series.append(timestamp, duration_ms)

        if timing_series != []:
            max_y = max(p.y() for s in timing_series for p in s.points())
        else:
            max_y = 0

        self.chart().createDefaultAxes()
        self.chart().axisX().setTitleText("Time (s)")
        self.chart().axisY().setTitleText("Elapsed time (ms)")
        self.chart().axisX().setRange(0, self._capture_len_s)
        self.chart().axisY().setRange(0, max(1.1 * max_y, 1))
        self.chart().update()


class _TimingMonitor(QWidget):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()
        plot = _TimingPlot(app_model)
        reset_button = QPushButton("Reset")
        reset_button.clicked.connect(plot.reset)

        layout = QVBoxLayout()
        layout.addWidget(reset_button)
        layout.addWidget(plot)
        self.setLayout(layout)


class AppModelViewer(QTabWidget):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()
        self.addTab(_CurrentStateView(app_model), "Current State")
        self.addTab(_SignalLog(app_model), "Signal Log")
        self.addTab(_TimingMonitor(app_model), "Timing Monitor")
        self.setMinimumSize(500, 500)
